import os
from . import demo as demo
from typing import overload

class StaticMethods:
    def __init__(self, *args, **kwargs) -> None: ...
    @overload
    @staticmethod
    def overloaded_static_method(value: int) -> int: ...
    @overload
    @staticmethod
    def overloaded_static_method(value: float) -> float: ...
    @staticmethod
    def some_static_method(a: int, b: int) -> int: ...

class TestStruct:
    field_readwrite: int
    field_readwrite_docstring: int
    def __init__(self, *args, **kwargs) -> None: ...
    @property
    def field_readonly(self) -> int: ...

def func_incomplete_signature(*args, **kwargs): ...
def func_returning_optional() -> int | None: ...
def func_returning_pair() -> tuple[int, float]: ...
def func_returning_path() -> os.PathLike: ...
def func_returning_vector() -> list[float]: ...
